__all__ = (
    "Animator",
    "AnimKey",
)

from typing import Any, Dict, List, Optional, Sequence

import bpy
import numpy as np


class Animator:
    """
    Wrapper around Blender's animation API, providing control of keyframe type and
    handles.

    Usage example:

    .. code-block:: py

       anim = Animator(obj, "location", index=1)   # Y location
       anim.animate(0, 0)
       anim.animate(10, 10, handle="VECTOR")
       anim.animate(20, 20, type="EXTREME")
    """

    obj: bpy.types.Object
    """Blender object to animate."""

    def __init__(self, obj: bpy.types.Object, data_path: str, index: int = 0,
            action_name: Optional[str] = None):
        """
        :param index: If a vector property, index of the component to animate, e.g. index 1 for Y location.
        :param action_name: Creates new action for obj with this name if not already present.
            Default obj.action
        """
        self.obj = obj
        self.index = index

        if obj.animation_data is None:
            obj.animation_data_create()
        if obj.animation_data.action is None:
            name = action_name if action_name is not None else f"{obj.name}.action"
            obj.animation_data.action = bpy.data.actions.new(name)

        self._fcurve = obj.animation_data.action.fcurves.find(data_path, index=index)
        if self._fcurve is None:
            self._fcurve = obj.animation_data.action.fcurves.new(data_path, index=index)

    @property
    def data_path(self):
        return self._fcurve.data_path

    def animate(self, frame: float, value: Any, handle: str = "AUTO_CLAMPED", type: str = "KEYFRAME"):
        """
        Insert keyframe.
        """
        key = self._fcurve.keyframe_points.insert(frame, value)
        key.handle_right_type = handle
        key.handle_left_type = handle
        key.type = type

        self._fcurve.update()


class AnimKey:
    """
    Set of value presets for animators.

    Each key is a list of values corresponding to each animator. The basis
    key is always defined at initialize. Other keys are diffs with respect to
    the basis key.

    Animation is done by assigning weights to each key.

    Usage example:

    .. code-block:: python

       anim1 = bmusic.Animator(obj, "location", 0)
       anim2 = bmusic.Animator(obj, "location", 1)
       anim3 = bmusic.Animator(obj, "location", 2)

       key = bmusic.AnimKey([anim1, anim2, anim3], [0, 0, 0])
       key["up"] = (0, 0, 1)
       key["cool"] = (1, 1, -1)
       key.animate(0)
       key.animate(30, up=2, handle="VECTOR", type="JITTER")
       key.animate(60, up=1, cool=2, type="EXTREME")
    """

    _animators: List[Animator]
    _keys: Dict[str, np.ndarray]

    def __init__(self, animators: List[Animator], basis: Sequence[float]):
        self._animators = animators
        self._keys = {}
        self._keys["basis"] = np.array(basis)

    def __iter__(self):
        return iter(self._keys)

    def __getitem__(self, key: str) -> np.ndarray:
        return self._keys[key]

    def __setitem__(self, key: str, value: Sequence[float]):
        """
        Key is converted to a difference with respect to basis
        i.e. (value - basis)
        """
        assert key not in ("basis", "handle", "type", "frame")
        assert len(value) == len(self._animators)
        value = np.array(value) - self["basis"]
        self._keys[key] = value

    def __delitem__(self, key: str):
        del self._keys[key]

    def __contains__(self, key: str):
        return key in self._keys

    def __len__(self):
        return len(self._keys)

    def __call__(self, *args, **kwargs):
        """
        Same as self.animate(...)
        Included for brevity.
        """
        return self.animate(*args, **kwargs)

    def animate(self, frame: float, handle: str = "AUTO_CLAMPED", type: str = "KEYFRAME",
            **kwargs):
        """
        Animate all animators.

        :param kwargs: Weight of each shape key.
        """
        value = self["basis"]
        for key, weight in kwargs.items():
            value = value + self[key]*weight

        for i, anim in enumerate(self._animators):
            anim.animate(frame, value[i], handle=handle, type=type)



__all__ = (
    "Scheduling",
)

from copy import deepcopy
from math import tanh
from typing import Callable

import bpy
import numpy as np
from tqdm import tqdm

from ..anim import AnimKey
from ..midi import MessageList
from .procedure import Procedure


class Scheduling(Procedure):
    """
    Schedule limited number of objects to switch between more notes to play.
    This procedure only moves the objects, but does not play the notes.
    You may want to combine it with something else e.g. Hammer.

    Parameters:

    animkeys
        List of animation keys, each corresponding to a hammer.

            - ``move0``, ``move1``, ``move2``, ...: Move to note index i.

    dist_f
        Function to get distance or cost between two note indexes.

    idle_time
        Time (sec) of pause before moving on to next note.

    depth
        Depth of search.
    """

    animkeys: list[AnimKey]
    dist_f: Callable[[int, int], float] = lambda i, j: abs(i-j)
    idle_time: float = 0.1
    depth: int = 3

    def animate(self):
        """
        Animates movement of each object to their messages.
        Returns which messages were scheduled to each object.

        :return: A list. Each element is a :class:`bmusic.MessageList` obj for the corresponding
            hammer to play.
        """
        self.fps = bpy.context.scene.render.fps  # For performance
        idle_time = self.idle_time * self.fps / 2
        notes_used = self.midi.notes_used

        # (note_ind, last_play_time)
        status = [[None, -1] for _ in self.animkeys]
        schedule = [[] for _ in self.animkeys]

        # Schedule notes
        print("BMusic: Scheduling: Scheduling notes...")
        min_reward = float("inf")
        for i, note in tqdm(enumerate(self.midi), total=len(self.midi.notes)):
            index, reward = self.best_choice(self.midi.notes, i, status, depth=self.depth)

            min_reward = min(min_reward, reward)
            schedule[index].append(note)
            status[index][0] = note.ind
            status[index][1] = max(note.start, status[index][1])

        print(f"BMusic: Scheduling: min_reward={min_reward}")

        # Animate motion
        print("BMusic: Scheduling: Animate motion...")
        midis = list(map(Midi.from_notes, schedule))
        for i, mid in enumerate(midis):
            for note in mid:
                prev = note.prev_start
                next = note.next_start
                frames = [note.start]

                """
                thres = idle_time * 2.5   # Don't want too jarring.
                if note.start-prev > thres:
                    frames.append(note.start-idle_time)
                if next-note.start > thres:
                    frames.append(note.start+idle_time)
                """

                # Midi changed so can't use note.ind
                ind = notes_used.index(note.note)
                kwargs = {f"move{ind}": 1}
                for f in frames:
                    self.animkeys[i].animate(f, **kwargs)

        return midis

    def best_choice(self, notes, note_i, status, depth=1):
        """
        Compute which hammer to choose.

        :return: (hammer_i, reward)
        """
        assert depth >= 1

        count = len(status)
        note = notes[note_i]

        some_not_played = None in [s[0] for s in status]

        reward = []
        for i in range(count):
            if some_not_played:
                # First make sure each hammer plays
                rew = 1e6 if status[i][0] is None else 0
                reward.append(rew)
                continue

            dist = self.dist_f(note.ind, status[i][0])
            dist = max(dist, 0.01)   # Divide by zero protection
            time = abs(note.start - status[i][1]) / self.fps
            rew = tanh(30 * time / dist)

            if depth > 1 and note_i < len(notes)-1:
                new_status = deepcopy(status)
                new_status[i][0] = note.ind
                new_status[i][1] = note.start
                _, depth_rew = self.best_choice(notes, note_i+1, new_status, depth-1)

                rew += depth_rew

            reward.append(rew)

        index = np.argmax(reward)
        rew = reward[index]

        return index, rew



__all__ = (
    "EXPONENTIAL",
    "LINEAR",
    "OSCILLATE",
)

import numpy as np

from .midi import *


def EXPONENTIAL(fac, t) -> float:
    """
    Exponential decay with factor ``fac`` (in units per second).
    """
    return np.exp(-fac * t)

def LINEAR(fac, t) -> float:
    """
    Linear decay with factor ``fac`` (in units per second).
    """
    return 1 - fac*t

def OSCILLATE(period, t) -> float:
    """
    Oscillate between -1 and 1 with period ``period`` (in seconds).
    When t=0, returns 1.
    """
    return np.cos(2*np.pi * t / period)



"""
Utilities for computing and representing affixes.

See docs for more info.
"""

__all__ = (
    "AffixMessage",
    "compute_affixes",
)

import numpy as np

from .midi import *


class AffixMessage(Message):
    """
    Prefix and suffix extension.

    ``prefix`` and ``suffix`` are lengths in frames.
    Based on application, suffix is either time after **start**
    or **end**.
    """

    prefix: float
    suffix: float

    def __repr__(self) -> str:
        return f"AffixMessage(msg={super().__repr__()}, prefix={self.prefix}, suffix={self.suffix})"


def compute_affixes(
        track: MessageList,
        split: float = 0.5,
        max_prefix: float = 1e9,
        max_suffix: float = 1e9,
        suffix_after_end: bool = False,
        hard_end: bool = False
        ) -> list[AffixMessage]:
    """
    For each message (taking into account neighboring messages in the track),
    compute length of prefix and suffix.

    For example, a prefix may be a hammer anticipating, and a suffix may be the
    bounce back.

    Prefixes and suffixes don't overlap with the next message --- they will be
    shortened if next/prev note is too close.

    The result may look like this:

    .. code-block:: none

       Messages:    |--------|-----------|--|--|--------|
       Return:    <<|>>>---<<|>>>------<<|><|><|>>>---<<|>>>

    Each vertical line represents when each message plays. On the line marked
    ``Return``, ``<`` shows the duration of prefix and ``>`` for suffix.

    Notice how the max prefix length is 2, and the max suffix length is 3.

    Also notice how the prefix and suffix don't overlap with the next message
    --- they are shortened when necessary.

    If we take this result and interpret prefixes and suffixes as hammer actions
    as described above, this would create a smooth-looking animation.

    See the docs page for more details.

    :param track: Track to process.
    :param split: Ratio between prefix and suffix when they overlap.
        ``0`` means completely prefix, and ``1`` means completely suffix.
    :param max_prefix: Maximum length of prefix in frames.
    :param max_suffix: Maximum length of suffix in frames.
    :param suffix_after_end: If true, suffix is after message ends. Else,
        suffix is immediately after message starts.
    :param hard_end: If true, cut off suffix when message ends (e.g. stop
        glowing when note is released). Otherwise, extend until ``max_suffix``
        (e.g. keep glowing indefinitely). Naturally, if used with
        ``suffix_after_end``, suffix will be 0.
    :return: List of Affix
    """
    def get_end(msg, fallback):
        """
        Get suffix starting point of message;
        returns msg.end if suffix_after_end is True,
        else msg.start
        """
        if msg is None:
            return fallback
        return msg.end if suffix_after_end else msg.start

    def get_msg_pair_info(msg1, msg2):
        """
        Returns convenient info about timestamps of prefix, overlap, etc.

        ``msg1`` and ``msg2`` should be two consecutive messages, where
        msg1 <= msg2.

        Return:
            nominal_suffix: msg1's suffix timestamp, disregarding overlap.
            nominal_prefix: msg2's prefix timestamp, disregarding overlap.
        """
        suffix = get_end(msg1, -1e9) + max_suffix
        prefix = msg2.start - max_prefix if msg2 is not None else 1e9
        return {
            "nominal_suffix": suffix,
            "nominal_prefix": prefix,
            "msg1": get_end(msg1, -1e9),
            "msg2": msg2.start if msg2 is not None else 1e9,
        }

    ret = []

    for msg in track:
        # Compute prefix
        info = get_msg_pair_info(msg.prev(), msg)
        if info["nominal_suffix"] > info["nominal_prefix"]:
            split_point = np.interp(split, [0, 1], [info["nominal_prefix"], info["nominal_suffix"]])
            split_point = np.clip(split_point, info["msg1"], info["msg2"])
            prefix_time = msg.start - split_point
        else:
            prefix_time = max_prefix

        # Compute suffix
        info = get_msg_pair_info(msg, msg.next())
        if info["nominal_suffix"] > info["nominal_prefix"]:
            split_point = np.interp(split, [0, 1], [info["nominal_prefix"], info["nominal_suffix"]])
            split_point = np.clip(split_point, info["msg1"], info["msg2"])
            suffix_time = split_point - get_end(msg, msg.start)
        else:
            suffix_time = max_suffix

        # Adjust suffix if hard_end is True
        if hard_end:
            if suffix_after_end:
                suffix_time = 0
            else:
                suffix_time = min(suffix_time, msg.end - msg.start)

        # Add to list
        affix = AffixMessage(msg.note, msg.velocity, msg.start, msg.end)
        affix.prefix = prefix_time
        affix.suffix = suffix_time
        ret.append(affix)

    return ret



"""
Procedures relating to intensity of something.
e.g. light, piano key's depression, string's vibration, etc.
"""

__all__ = (
    "Intensity",
    "IntensityOnOff",
    "IntensityFade",
)

from typing import Callable

import bpy

from ..affix import *
from ..anim import *
from ..midi import *
from ..utils import *
from .procedure import ForEachProc


class Intensity(ForEachProc):
    """
    Base intensity procedure.

    Parameters:

    animkey
        Animation key with the following keys:

            - basis: Resting (intensity 0) position.
            - on: Playing (intensity max) position.
    """

    animkey: AnimKey


class IntensityOnOff(Intensity):
    """
    Turns on when a note starts and off when it ends.

    :Keyframe types:

        - JITTER: Resting (intensity = 0).
        - BREAKDOWN: Playing.

    Parameters:

    duration
        Time, in seconds, to spend interpolating between states.

    vector_handles
        Whether to use vector handles (no easing in or out).
    """

    duration: float = 0.05
    vector_handles: bool = False

    def animate(self):
        duration = self.duration * bpy.context.scene.render.fps
        handle = "VECTOR" if self.vector_handles else "AUTO_CLAMPED"

        msgs = compute_affixes(self.midi, max_prefix=duration, max_suffix=duration, suffix_after_end=True)
        for msg in msgs:
            intensity = self.get_intensity(msg)

            # For brevity, make a list and call func later.
            # List of (frame, value, key_type)
            keys = []

            # Initial resting position
            keys.append((msg.start-msg.prefix, 0, "JITTER"))
            # Playing through note
            keys.append((msg.start, intensity, "BREAKDOWN"))
            keys.append((msg.end, intensity, "BREAKDOWN"))
            # Resting after note
            if msg.suffix >= duration:
                # Only if suffix is full length; otherwise, the next note will take care of it.
                keys.append((msg.end+msg.suffix, 0, "JITTER"))

            for frame, value, type in keys:
                self.animkey.animate(frame, handle=handle, type=type, on=value)


class IntensityFade(Intensity):
    """
    Turns on when a note starts and fades away.

    :Keyframe types:

        - JITTER: Resting (intensity = 0).
        - BREAKDOWN: Fading.
        - EXTREME: Max intensity.

    Parameters:

    fade_func
        Function that curves linear time to fade intensity.
        Takes parameters ``(t,)`` (time in seconds since note start) and returns
        between 0 and 1, where 0 is off and 1 is max intensity. There are a few
        predefined functions ``bmusic.utils``

    start_time
        Time, in seconds, from off to initial on.

    key_interval
        Interval in seconds for decay keyframes. Avoids unneccessary
        keyframing on every frame.

    off_thres
        Threshold for intensity to be considered off. Keyframing will stop
        at this point. If using a custom ``fade_func`` that is NOT monotonically decreasing,
        set this to ``0`` (or negative).

    max_len
        Maximum length of note, in seconds. Keyframing will stop after this
        time, even if the note is still playing.

    note_end
        Whether to cut intensity to 0 at end of note. If false, the note will
        keep fading until next play, even if the note was released.
    """

    fade_func: Callable[[float], float] = lambda t: EXPONENTIAL(0.6, t)
    start_time: float = 0.05
    key_interval: float = 0.3
    off_thres: float = 0.001
    max_len: float = 60
    note_end: bool = True

    def animate(self):
        fps = bpy.context.scene.render.fps
        start_time = self.start_time * fps
        key_interval = self.key_interval * fps
        max_len = self.max_len * fps

        msgs = compute_affixes(self.midi, max_prefix=start_time, max_suffix=max_len, suffix_after_end=False, split=0)
        last_value = 0
        for msg in msgs:
            intensity = self.get_intensity(msg)

            # Start
            self.animkey.animate(msg.start-msg.prefix, on=last_value, handle="VECTOR", type="BREAKDOWN")
            self.animkey.animate(msg.start, on=intensity, handle="VECTOR", type="EXTREME")

            # Fading
            frame = msg.start
            added_key = False
            while True:
                frame += key_interval
                if frame > msg.start + msg.suffix:
                    break
                added_key = True

                last_value = intensity * self.fade_func((frame - msg.start) / fps)

                if self.off_thres > 0 and last_value < self.off_thres:
                    # Keyframe off
                    self.animkey.animate(frame, on=0, handle="VECTOR", type="JITTER")
                    last_value = 0
                    break

                self.animkey.animate(frame, on=last_value, handle="VECTOR", type="BREAKDOWN")

            if not added_key:
                # Suffix was too short for ``key_interval``.
                # Here, we set ``last_value`` to the appropriate value, and the next iteration will key it for us.
                last_value = intensity * self.fade_func(msg.suffix / fps)



__all__ = (
    "Hammer",
)

import bpy

from ..affix import compute_affixes
from ..anim import AnimKey
from ..utils import *
from .procedure import ForEachProc


class Hammer(ForEachProc):
    """
    Hammer movement: Resting, preparing, hitting, recoil, wobbling.

    :Keyframe types:

        - JITTER: Resting.
        - BREAKDOWN: Preparing to hit.
        - EXTREME: Hitting.
        - KEYFRAME: Recoil and oscillating.

    Parameters:

    animkey
        Animation key:

            - hit: Hitting.
            - prepare: Preparing to hit.
            - recoil: Bounce back after hit.

    prepare_time
        Duration (sec) of rest to prepare.

    hit_time
        Duration (sec) of prepare to hit movement.

    recoil_time
        Duration (sec) of hit to recoil movement.

    osc_period
        Duration (sec) of each oscillation after hit.

    osc_count
        Number of oscillations to perform.

    osc_decay
        Factor by which oscillation intensity decays each time.
    """

    animkey: AnimKey
    prepare_time: float = 0.15
    hit_time: float = 0.08
    recoil_time: float = 0.13
    osc_period: float = 0.35
    osc_count: int = 4
    osc_decay: float = 0.5

    def animate(self):
        fps = bpy.context.scene.render.fps
        hit_time = self.hit_time * fps
        prepare_time = self.prepare_time * fps
        recoil_time = self.recoil_time * fps
        osc_period = self.osc_period * fps
        osc_count = self.osc_count + 1   # for ending "rest" oscillation.
        osc_decay = self.osc_decay

        before_time = prepare_time + hit_time
        osc_time = osc_period * osc_count
        after_time = osc_time + recoil_time

        msgs = compute_affixes(self.midi, max_prefix=before_time, max_suffix=after_time, split=0)
        prev_ended = True
        for msg in msgs:
            if prev_ended:
                # Need to key at rest first.
                self.animkey.animate(msg.start - before_time, type="JITTER")

            # Prepare
            intensity = self.get_intensity(msg)
            self.animkey.animate(msg.start - msg.prefix/2, type="BREAKDOWN", prepare=intensity)

            # Hit
            self.animkey.animate(msg.start, type="EXTREME", handle="VECTOR", hit=1)

            # Oscillation
            prev_ended = False
            frame = recoil_time
            for i in range(osc_count):
                if frame >= msg.suffix:
                    break

                if i == osc_count-1:
                    value = 0
                else:
                    value = intensity * osc_decay ** (i+1)
                name = "prepare" if i % 2 == 0 else "hit"
                kwargs = {name: value}
                self.animkey.animate(msg.start + frame, type="KEYFRAME", **kwargs)

                frame += osc_period

            # Ending
            if msg.suffix >= after_time:
                self.animkey.animate(msg.start + frame, type="JITTER")
                prev_ended = True



__all__ = (
    "ProcMetaCls",
    "Procedure",
    "MusicProc",
    "ForEachProc",
)

import numpy as np

from ..midi import *


def _dummy():
    pass
function = type(_dummy)


class ProcMetaCls(type):
    """
    Metaclass for procedures.
    Automatically generates property documentation.

    Sets ``cls._params``, type set.

    In your class's docstring:

    1. Write a parameters section at the end; this is denoted with the text ``Parameters:``.
       See an existing class for an example.
    2. For each parameter, document it.
       Again, see an existing example.

    Example (put this at the end of docstring):

    .. code-block:: none

        Parameters:

        param1
            This is a parameter.
            Multiline descriptions are allowed.
            Must be 4 spaces indent.

        param2
            ...
    """

    def __new__(cls, name, bases, attrs):
        params, base_params = cls.get_params(bases, attrs)
        attrs["_params"] = params | base_params
        annotations = attrs.get("__annotations__", {})

        if attrs.get("_gen_docs", True):
            # Get and indent docstring.
            docstr = attrs.get("__doc__")
            if not isinstance(docstr, str):
                docstr = ""
            docstr = cls.reindent(docstr, 0).strip()

            # Extract parameter docs.
            param_docs = {}
            if "Parameters:" in docstr:
                ind = docstr.find("Parameters:")
                param_docstr = docstr[ind:].strip()
                docstr = docstr[:ind].strip()

                curr_param = None
                for i, line in enumerate(param_docstr.split("\n")):
                    # Skip first line.
                    if i == 0:
                        continue

                    if line.startswith(" ") and curr_param is not None:
                        param_docs[curr_param] += line[4:] + "\n"
                    else:
                        if line.strip():
                            curr_param = line.strip()
                            param_docs[curr_param] = ""

            # Add docs, type hints, and defaults.
            docstr += "\n\n:Parameters:\n\n"
            all_params = sorted(params) + sorted(base_params)
            if all_params:
                for param in all_params:
                    docstr += f"- ``{param}``"
                    if param in base_params:
                        docstr += " *(inherited)*"
                    else:
                        docstr += ":" + cls.reindent(param_docs.get(param, "").strip(), 2) + "\n\n"
                        if param in annotations:
                            docstr += f"  - **Type:** {annotations[param].__name__}\n"
                        if param in attrs:
                            docstr += f"  - **Default:** {attrs[param]}\n\n"
                    docstr += "\n"
            else:
                docstr += "None\n\n"

            attrs["__doc__"] = docstr

        return super().__new__(cls, name, bases, attrs)

    @staticmethod
    def get_params(bases, attrs) -> tuple[set[str], set[str]]:
        """
        Returns set of procedure parameters, extending from bases.
        """
        params = set()
        for key, value in attrs.items():
            if key.startswith("_"):
                continue
            if isinstance(value, (property, function)):
                continue
            params.add(key)
        if "__annotations__" in attrs:
            for key, value in attrs["__annotations__"].items():
                if key.startswith("_"):
                    continue
                if key in params:
                    continue
                params.add(key)

        base_params = set()
        for cls in bases:
            if hasattr(cls, "_params"):
                base_params.update(cls._params)

        return params, base_params

    @staticmethod
    def reindent(text: str, indent: int) -> str:
        max_indent = None
        for line in text.split("\n"):
            if not line.strip():
                continue
            curr_indent = len(line) - len(line.lstrip())
            if max_indent is None:
                max_indent = curr_indent
            else:
                max_indent = min(max_indent, curr_indent)
        if max_indent is None:
            return text

        lines = []
        for line in text.split("\n"):
            lines.append(" "*indent + line[max_indent:])
        return "\n".join(lines)



class Procedure(metaclass=ProcMetaCls):
    """
    Procedure base class.

    A procedure animates objects, with input parameters from the user.
    Procedures are generalized: For example, a hammer procedure can be applied
    to a drum mallet, a piano key, etc.

    Extend from this class to create your animator. Subclasses inherit
    available parameters from the parent class.

    This class uses the :class:`bmusic.ProcMetaCls` metaclass to automatically
    generate property docs.

    - For each property, type hint and/or define a default value in the class.
    - Then write the docstring in the class's docstring, explained further in the docs.
    - If you don't want automatic docs, set ``_gen_docs = False``.
    """

    _gen_docs = True
    _params: set[str] = set()
    """This is set by metaclass"""

    def __init__(self, **kwargs):
        """
        Don't override this method; this automatically sets parameters from kwargs.

        You can define the ``init`` method, which will be called from here.
        ``init`` is passed the same kwargs as this method.
        """
        for key, value in kwargs.items():
            if key in self._params:
                setattr(self, key, value)
        for param in self._params:
            if not hasattr(self, param):
                raise ValueError(f"Missing parameter: {param}")

        self.init(**kwargs)

    def init(self, **kwargs):
        """
        This is called after parameters are set from ``__init__``.

        Override this method to do custom initialization.
        """
        pass

    def animate(self):
        """
        Do all animation.
        """
        raise NotImplementedError("Subclass must implement animate()")


class MusicProc(Procedure):
    """
    Procedure superclass with utilities for music animation.

    Yes, this whole library is meant for music animation, but the base Procedure class
    is empty.
    This class has some stuff specific for music animation.

    Parameters:

    midi
        :class:`bmusic.MessageList` object containing messages to animate.
    """

    midi: MessageList


class ForEachProc(MusicProc):
    """
    Base class meant for animating the same thing on each message.
    e.g. for each message, do hammer; light up; etc.

    Intensity of message is interpolated from velocity, using given bounds.

    Parameters:

    min_intensity
        Minimum peak intensity. Happens when velocity is 0.

    max_intensity
        Maximum peak intensity. Happens when velocity is 127.
    """

    min_intensity: float = 0
    max_intensity: float = 1

    def get_intensity(self, msg: Message):
        """
        Returns interpolated intensity from velocity.
        """
        return np.interp(msg.velocity, [0, 127], [self.min_intensity, self.max_intensity])



"""
"note" refers to a pitch, represented as an integer in MIDI.
For example, there are 88 notes on a piano.

"message" refers to an instance of a note being played.
For example, this MIDI file has 100 messages,
using a total of 10 different notes.
"""

from typing import Generator, Optional, Sequence

import bpy
import mido

__all__ = (
    "Message",
    "MessageList",
    "parse_midi",
    "split_chords",
)


class Message:
    """
    Message with absolute start and end time in frames.
    Also supports linked-list like behavior within track.
    """

    note: int
    velocity: int
    start: float
    end: float

    def __init__(self, note, velocity, start, end, msglist: Optional["MessageList"] = None, index: Optional[int] = None):
        """
        :param msglist: MessageList object this message belongs to.
        :param index: Index of this message in the track. Allows for linked-list behavior.
        """
        self.note = note
        self.velocity = velocity
        self.start = start
        self.end = end

        self.msglist = msglist
        self.index = index

    def __repr__(self):
        return (f"bmusic.Message(note={self.note}, velocity={self.velocity}, "
                f"start={self.start}, end={self.end})")

    def copy(self) -> "Message":
        return Message(self.note, self.velocity, self.start, self.end, self.msglist, self.index)

    def diff(self, other: "Message") -> float:
        """
        Calculate difference in frames between this message and another's start.
        Positive if this message starts after other.
        """
        return self.start - other.start

    def next(self) -> Optional["Message"]:
        """
        Get next message in msglist.
        """
        assert self.index is not None
        assert self.msglist is not None

        ind = self.index + 1
        if ind >= len(self.msglist._messages):
            return None
        return self.msglist._messages[ind]

    def next_start(self, fallback: float = 1e9) -> float:
        """
        Timestamp of next message's start.

        :param fallback: Value to return if no next message.
        """
        return self.next().start if self.next() else fallback

    def prev(self) -> Optional["Message"]:
        """
        Get previous message in msglist.
        """
        assert self.index is not None
        assert self.msglist is not None

        ind = self.index - 1
        if ind < 0:
            return None
        return self.msglist._messages[ind]

    def prev_start(self, fallback: float = -1e9) -> float:
        """
        Timestamp of previous message's start.

        :param fallback: Value to return if no previous message.
        """
        return self.prev().start if self.prev() else fallback

    def prev_end(self, fallback: float = -1e9) -> float:
        """
        Timestamp of previous message's end.

        :param fallback: Value to return if no previous message.
        """
        return self.prev().end if self.prev() else fallback


class MessageList:
    """
    Sequence of messages sorted by start time.
    """

    def __init__(self, messages: Optional[Sequence[Message]] = None):
        """
        Initialize from message.
        Creates copies of each message.
        Also associates each message with self.
        """
        if messages is None:
            messages = []

        self._messages = []
        for msg in messages:
            msg = msg.copy()
            msg.msglist = self
            msg.index = len(self._messages)
            self._messages.append(msg)

        self._messages.sort(key=lambda n: n.start)
        self._noteset = sorted(set(n.note for n in self._messages))

    def __iter__(self) -> Generator[Message, None, None]:
        yield from self._messages

    def __len__(self):
        return len(self._messages)

    def __getitem__(self, item):
        return self._messages[item]

    def __repr__(self):
        return f"bmusic.MessageList({self._messages})"

    def noteset(self):
        """
        Set of all notes used in this track.
        """
        return self._noteset

    def duration(self) -> float:
        """
        Duration, in frames, between first msg's start and last msg's end.
        """
        return self._messages[-1].end - self._messages[0].start
    
    def filter_notes(self, notes: Sequence[int]) -> "MessageList":
        """
        Only keep messages with note in the given list.
        """
        return MessageList([n for n in self._messages if n.note in notes])

    def split_notes(self):
        """
        Split into multiple tracks, each containing only one note.
        """
        for note in self.noteset():
            yield MessageList([n for n in self._messages if n.note == note])


def parse_midi(path: str, offset: float = 0, fps: Optional[float] = None) -> MessageList:
    """
    Parse MIDI from a file.

    :param path: Path to midi file.
    :param offset: Duration in frames of first message's start.
    :param fps: Frames per second (None for blender fps)
    """
    notes = []

    if fps is None:
        fps = bpy.context.scene.render.fps
    midi = mido.MidiFile(path)

    # State of the last key press of note i.
    starts = [0] * 1000
    vels = [0] * 1000

    frame = 0
    started = False
    for msg in midi:
        if started:
            frame += msg.time * fps

        if msg.type.startswith("note_"):
            started = True
            note = msg.note
            vel = msg.velocity if msg.type == "note_on" else 0
            if vel == 0:
                n = Message(note, vels[note], starts[note]+offset, frame+offset)
                notes.append(n)
            else:
                starts[note] = frame
                vels[note] = vel

    return MessageList(notes)


def split_chords(midi: MessageList, threshold: float) -> list[list[Message]]:
    """
    Split a MIDI track into chords --- messages that play roughly at the same time.
    Messages less than threshold apart will be combined into one chord.
    Threshold units are whatever midi units are. Most likely frames.
    """
    chords = []

    last_time = None
    chord = []
    for msg in midi:
        if last_time is None:
            chord.append(msg)
        elif abs(msg.start - last_time) < threshold:
            chord.append(msg)
        else:
            chords.append(chord)
            chord = [msg]

    return chords
